/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "ssfem_kernel.h"
#include "flxrandomFEM.h"

using namespace std;

RBRV_set_box* flxSFEM::rbrv_box = NULL;

#ifdef __WINDOWS__
  #pragma comment(lib, "flxfem")
#endif


// -------------------------- PolyChaos ----------------------------------------------------------------------

PolyChaos::PolyChaos(const int MV, const int pV):qS(NULL),H(NULL),ExpSq(NULL)
{
  GenerateSequence(MV,pV);
}

PolyChaos::~PolyChaos()
{
  delMem();
}

int* PolyChaos::ExtrSeq(const int n, int *balls) const
{
  int *a = new int[M];
  a[0]=balls[0];
  for (int i = 1; i<M-1;i++) {
    a[i] = balls[i]-balls[i-1]-1;
  }
  a[M-1]=n-balls[M-2]-1;
  return a;
}

const bool PolyChaos::NxtSmp(const int n, const int m, int* balls) const
{
  for (int i=0;i<m;i++) {
    if (balls[i]!=i) {
      balls[i]-=1;
      for (int j=i-1;j>=0;j--) {
	balls[j]=balls[j+1]-1;
      }
      return true;
    }
  }
  return false;
}

void PolyChaos::GenerateSequence(const int MV, const int pV)
{

  // Delete Sequences
  // ================
  delMem();
  M = MV;
  p=pV;

  // Construction of 1D Hermite Polynomials
  // ======================================
  H = new tdouble* [p+1];
  H[0] = new tdouble [1];
  H[1] = new tdouble [1];
  H[0][0] = ONE;
  H[1][0] = ONE;
  for (int i = 2; i <= p; i++) {
    int t = (i+2)/2;	// number of coefficients unequal to 0
    int g = (i+1) % 2;	// even number=1; uneven number=0
    H[i] = new tdouble [t];
    // Construction-part of previous polynomial
    int tu = (i+1)/2;	// number of coefficients of the previous polynomial
    H[i][0]=0;
    for (int j = 0; j < tu; j++) {
      H[i][j+g]=H[i-1][j];
    }
    // Construction-part of last but one
    for (int j = 0; j < t-1; j++) {
      H[i][j] += (ONE-tdouble(i))*H[i-2][j];
    }
  }
  
//   // DEBUG-OUTPUT
//   for (int i = 0; i<=p; i++) {
//     cout << i << ".: ";
//     for (int j = 0; j<(i+2)/2;j++){
// 	cout << H[i][j] << ", ";	
//     }
//     cout << endl;
//   }

// Construction of Polynomial Basis
// ================================

  jA = 0;
  for (int q=0;q<=p;q++) {
    jA+=tulong(BinomCoeff(M+q-1,M-1));
  }
//   cout << jA << endl;
  qS = new int*[jA+1];
    for (size_t i=0;i<=jA;++i) qS[i] = NULL;
  if (M==1) {
    for (int i=0;i<int(jA);i++) {
      qS[i] = new int [1];
      qS[i][0]=i;
    }
  } else {
    tulong pos=0;
    for (int q=0;q<=p;q++) {
      int n = M+q-1;
      int m = M-1;
      int *balls = new int[m];
      for (int i = 0; i<m; i++)
      {
	balls[i]=q+i;
      }
      do {
	qS[pos] = ExtrSeq(n,balls); 
	pos++;
      } while (NxtSmp(n,m,balls));
      delete [] balls;
    }
  }
  
//   // DEBUG-OUTPUT
//   cout << "Polynomial Basis" << endl;
//   for (int i = 0; i<jA;i++) {
//     for (int j=0; j<M;j++) {
//       cout << qS[i][j] << " " ;
//     }
//     cout << endl;
//   }

  ExpSq = new tdouble[jA];
  for (tulong i = 0; i<jA;i++) {
    ExpSq[i] = calc_ExpPB2(i);
  }
}

void PolyChaos::delMem()
{
  if (qS!=NULL) {
    for (tulong i=0;i<=jA;i++) {
      if (qS[i]) delete [] qS[i];
    }
    delete [] qS;
    qS=NULL;
  }
  if (H!=NULL) {
    for (int i = 0; i <= p; i++) {
      delete [] H[i];
    }
    delete [] H;
    H=NULL;
  }
  if (ExpSq!=NULL) {
    delete[] ExpSq;
    ExpSq=NULL;
  }
  M=0;
  p=0;
  jA=0;
}

const tdouble PolyChaos::get_ExpPB2(const tulong j) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_ExpPB2_1", ssV_2.str() );
    }
  #endif
  return ExpSq[j];
}

FlxPeval* PolyChaos::get_H1D(const int q, const int m) const
{
  #if FLX_DEBUG
    if ( q < 0 || q > p) {
      ostringstream ssV_2;
      ssV_2 << "q (q=" << q << "; q.th Hermite Polynomail) has to be in the Range [0;" << p << "].";
      throw FlxException("PolyChaos::get_H1D_1", ssV_2.str() );
    }
    if ( m < 1 || m > M) {
      ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("PolyChaos::get_H1D_2", ssV_2.str() );
    }
  #endif
  FlxPeval* hm = new FlxPeval(M);
  int t = (q+2)/2;	// number of coefficients unequal to 0
  int u = (q) % 2;	// even number=0; uneven number=1
  for (int i=0; i<t;i++) {
    FlxPeval_term tp(M,m,i*2+u);
    hm->add_term(tp,H[q][i]);
  }
  return hm;
}

FlxPeval* PolyChaos::get_Poly(const tulong j) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("PolyChaos::get_Poly_1", ssV_2.str() );
    }
  #endif
  FlxPeval* hm = new FlxPeval(M,true);
  FlxPeval *hmT1; FlxPeval *hmT2;
  int *qSj = qS[j];
  for (int i = 0; i < M; ++i) {
    if (qSj[i]!=0) {
      hmT1 = hm;
      hmT2 = get_H1D(qSj[i],i+1);
      hm = new FlxPeval(*hmT1,*hmT2);
      delete hmT1;
      delete hmT2;
    }
  }
  return hm;  
}

const tdouble PolyChaos::calc_Realization(const flxVec& U, const flxVec& x) const
{
  #if FLX_DEBUG
    if ( U.get_N() != jA ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Realization_1", ssV_2.str() );
    }
    if (x.get_N() != static_cast<tuint>(M)) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Realization_2", ssV_2.str() );
    }
  #endif
  tdouble u = U[0];
  for (tulong i = 1; i < jA; ++i) {
    u+= U[i]*calc_Poly(i,x);
  }
  return u;  
}

FlxPeval* PolyChaos::get_Realization(const flxVec& U) const
{
  FlxPeval* sp = new FlxPeval(M);
  sp->add_term(FlxPeval_term(M),U[0]);
  
  FlxPeval* spT;
  for (tulong i = 1; i < jA; ++i) {
    spT = get_Poly(i);
    spT->operator*=(U[i]);
    sp->operator+=(*spT);
    delete spT;
  }
  return sp; 
}

const tdouble PolyChaos::calc_ExpPB2(const tulong j) 
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_ExpPB2_1", ssV_2.str() );
    }
  #endif
  tdouble E = 1;
  for (int i=0;i<M;i++) {
    E *= Factorial(qS[j][i]);
  }
  return E;
}

const tdouble PolyChaos::calc_c_ijk(const int i, tulong j, tulong k)  // TODO: inline?
{
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_2", ssV_2.str() );
    }
    if ( k < 0 || k >= jA) {
      ostringstream ssV_2;
      ssV_2 << "k (k=" << k << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_2", ssV_2.str() );
    }
    if ( i < 0 || i > M) {
      ostringstream ssV_2;
      ssV_2 << "i (i=" << i << "; has to be in the Range [0;" << M << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_3", ssV_2.str() );
    }
  #endif
  
  // i = 0 has to be treated separately
  if (i == 0) {
    if (j != k) return ZERO;
    else return ExpSq[j];
  }
  
  if (j==k) return ZERO;
  // Consider orthogonality
    for (int u=0;u<M;u++) {
      if (u!=i-1) {
	if (qS[j][u]!=qS[k][u]) return ZERO;
      }
    }
    int vp = qS[j][i-1];
    int vq = qS[k][i-1];
  // make sure vp > vq
    if (vp<vq) {
      std::swap(vp,vq);
      std::swap(j,k);
    }
    if (vp-1!=vq) return ZERO;
  // Compute c_ijk
    return ExpSq[j];
}

void PolyChaos::calc_Kjk(tulong j, tulong k, int& i, tdouble& f)
{
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_Kjk_1", ssV_2.str() );
    }
    if ( k < 0 || k >= jA) {
      ostringstream ssV_2;
      ssV_2 << "k (k=" << k << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_Kjk_2", ssV_2.str() );
    }
  #endif
  
  if ( j == k) {
    i = 0;
    f = ExpSq[j];
    return;
  }
  
  int *qS1 = qS[j];
  int *qS2 = qS[k];
  
  bool hit = false;
  int ihit = 0;
  for (int m = 0; m < M; m++) {
    if (qS1[m] != qS2[m]) {
      if (hit) {
	i = 0; 
	f = 0;
	return;
      } else {
	hit = true;
	ihit = m;
      }
    }
  }
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( !hit ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Kjk_3", ssV_2.str() );
    }
  #endif
  
  int vp = qS1[ihit];
  int vq = qS2[ihit];
  
  // make sure vp > vq
    if (vp<vq) {
      std::swap(vp,vq);
      std::swap(j,k);
    }
    if (vp-1!=vq) {
      i = 0; 
      f = 0;
      return;
    } else {
      i = ihit+1;
      f = ExpSq[j];
      return;
    }
}


// -------------------------- flxStiffnessMtxSFEM -------------------------------------------------------

flxStiffnessMtxSFEM::flxStiffnessMtxSFEM(const int dim, const int order, flxFEM* FEM, std::vector< RBRV_set_base* > set_vec)
: flxStiffnessMtxBase(FEM),pc(dim,order),StfMtxV(NULL), RndBox(set_vec)
{

}

flxStiffnessMtxSFEM::~flxStiffnessMtxSFEM()
{
  delete_StfMtxV();
}

void flxStiffnessMtxSFEM::delete_StfMtxV()
{
  if (StfMtxV == NULL) return;
  int M = pc.get_dim();
  for (int i = 0; i <= M; ++i) {
    if (StfMtxV[i]) delete StfMtxV[i];
  }
  delete [] StfMtxV;
}

void flxStiffnessMtxSFEM::output_solVec(ostream& sout, bool verbose)
{
  if (SolutionVec==NULL) return;
  sout << "Solution of the system:" << std::endl;
  const tulong P = pc.get_SeqNumb();  
  flxVec sD(DofA);
  flxVec vtmp(DofA);
  for (tulong i = 1; i < P; ++i) {
    flxVec h1(&((*SolutionVec)[i*DofA]),DofA);
    vtmp.mult_coeff(h1,h1);
    sD.add(vtmp,pc.get_ExpPB2(i));
  }
  for (tnlong i = 0; i < DofA; ++i) {
    sout << i << " " << (*SolutionVec)[i] << " (" << sqrt(sD[i]) << ")" << std::endl;
  }
  if (verbose) {
    sout << (*SolutionVec) << std::endl;
  }
}

const tdouble flxStiffnessMtxSFEM::get_U(const tnlong dofindex) const
{
  if (SolutionVec==NULL) {
    std::ostringstream ssV;
    ssV << "Solution vector is empty.";
    throw FlxException("flxStiffnessMtxSFEM::get_U_1", ssV.str() );
  }
  if (dofindex >= DofA) return ZERO; 	// TODO predisplacement ??? FunNodeU::data
  if (current_RF_mode_get() < 0) {
    const tulong P = pc.get_SeqNumb();  
    flxVec Uvec(P);
    Uvec.slice(&((*SolutionVec)[dofindex]),DofA);
    // get vector of random realizations
      flxVec xV(RndBox.get_NRV());
      RndBox.get_x_Vec(xV.get_tmp_vptr());
    return pc.calc_Realization(Uvec,xV);
  } else {
    return (*SolutionVec)[current_RF_mode_get()*DofA+dofindex];
  }
}

void flxStiffnessMtxSFEM::get_U(const tnlong dofindex, tdouble* v, const tuint N) const
{
  for (tuint i=0;i<N;++i) {
    v[i] = get_U(dofindex+i);
  }
}

const tdouble flxStiffnessMtxSFEM::get_F(const flxDOFn* Dof)
{
  throw FlxException_NotImplemented("flxStiffnessMtxSFEM::get_F");
}

FlxPeval* flxStiffnessMtxSFEM::get_Pol(const tnlong dofindex)
{
  if (SolutionVec==NULL) {
    std::ostringstream ssV;
    ssV << "Solution vector is empty.";
    throw FlxException("flxStiffnessMtxSFEM::get_Pol_1", ssV.str() );
  }  
  if (dofindex >= DofA) return new FlxPeval(pc.get_dim()); 	// TODO predisplacement ??? FunNodeU::data
  
  const tulong P = pc.get_SeqNumb();  
  flxVec Uvec(P);
  Uvec.slice(&((*SolutionVec)[dofindex]),DofA);
  return pc.get_Realization(Uvec);
}

void flxStiffnessMtxSFEM::assemble_forc(bool errSerious)
{
  // TODO assemble force vector in case of random forces
  const tulong P = pc.get_SeqNumb();  
  if (ForceVec == NULL) ForceVec = new flxVec(P*DofA);
  
  flxVec* SSFEM_ForceVec = ForceVec;
  ForceVec = NULL;
  
  flxStiffnessMtxBase::assemble_forc();
  flxVec f_tmp(SSFEM_ForceVec->get_tmp_vptr(),tuint(DofA));
  f_tmp = *ForceVec;
  delete ForceVec;
  
  ForceVec = SSFEM_ForceVec;
}

const double flxStiffnessMtxSFEM::solve(tdouble& eps, tnlong& iter, tuint pcn, bool startZero, const bool errSerious, const bool verboseOutput)
{
  flxVec mtv(RndBox.get_NOX());
  RndBox.get_mean_Vec(mtv.get_tmp_vptr());
  RndBox.set_smp_x(mtv);
  return flxStiffnessMtxBase::solve(eps, iter, pcn, startZero, errSerious, verboseOutput);
}

void flxStiffnessMtxSFEM::allocate_SolutionVec()
{
  if (SolutionVec==NULL) {
    SolutionVec=new flxVec(pc.get_SeqNumb()*DofA);
  }
}

void flxStiffnessMtxSFEM::assemble_stf()
{
  if (preDispA>0) {
    throw FlxException_NotImplemented("flxStiffnessMtxSFEM::assemble_stf");
  }
  // First level of assembly
    const int cri = current_RF_mode_get();
    delete_StfMtxV();
    const int M = pc.get_dim();
    StfMtxV = new FlxMtx_baseS*[M+1];
      for (int i=0;i<=M;++i) StfMtxV[i] = NULL;
    for (int i = 0; i <= M; ++i) {
      current_RF_mode_set(i);
      flxStiffnessMtxBase::assemble_stf();
      StfMtxV[i] = StfMtx;
      StfMtx = NULL;
    }
    current_RF_mode_set(cri);
  // Second level of assembly
    tulong P = pc.get_SeqNumb();
    valarray<tdouble> fM((P*P+P)/2);
    valarray<int> KM((P*P+P)/2);
    int i;
    tdouble f;
    for (tulong j = 0, l=0; j < P; ++j) {
      for (tulong k = 0; k <= j; ++k,++l) {
	pc.calc_Kjk(j,k,i,f);
	if (f>ZERO) {
	  fM[l] = f;
	  KM[l] = i;
	}
      }
    }
    StfMtx = new FlxMtxSparsSFEMSym(fM,KM,StfMtxV,P);
}




// -------------------------- SFEM ----------------------------------------------------------------------


flxSFEM::flxSFEM(int order, FlxConstantBox* constBox)
 :flxFEM(constBox),p(order)
{
  
}

flxSFEM::~flxSFEM()
{

}

flxStiffnessMtxSFEM* flxSFEM::get_Sstfmtx()
{
  if (StiffnessMtx == NULL) {
    std::ostringstream ssV;
    ssV << "The stiffness matrix has not yet been assembled.";
    throw FlxException("flxSFEM::get_Sstfmtx_1", ssV.str() );
  }
  flxStiffnessMtxSFEM* sStfMtx = dynamic_cast<flxStiffnessMtxSFEM*> (StiffnessMtx);
  #if FLX_DEBUG
    if (sStfMtx == NULL) {
      std::ostringstream ssV;
      ssV << "This is not a stochastic finite element analysis.";
      throw FlxException("flxSFEM::get_Sstfmtx_2", ssV.str() );
    }
  #endif
  return sStfMtx;
}

FlxPeval* flxSFEM::get_Pol(const tnlong dofindex)
{
  flxStiffnessMtxSFEM* sStfMtx = get_Sstfmtx();
  return sStfMtx->get_Pol(dofindex);
}

RBRV_constructor& flxSFEM::get_RndBox()
{
  flxStiffnessMtxSFEM* sStfMtx = get_Sstfmtx();
  return sStfMtx->RndBox;
}

void flxSFEM::assembleDOF()
{
  // check for consistency
    std::vector<RBRV_set_base*> set_vec = rbrv_box->get_entire_set_vec();
    for (size_t i=set_vec.size();i>0;--i) {
      RBRV_set_noise_RF* setRF = dynamic_cast<RBRV_set_noise_RF*>(set_vec[i-1]);
      if (setRF==NULL) {
	set_vec.erase(set_vec.begin()+i-1);
      }
    }
    if (set_vec.size()==0) {
      std::ostringstream ssV;
      ssV << "There exist no random variables that the stochastic finite element module can handle.";
      throw FlxException("flxSFEM::assembleDOF_1", ssV.str() );
    }
    RBRV_constructor RndBox(set_vec);
    if (RndBox.get_NRV()==0) {
      std::ostringstream ssV;
      ssV << "There exist no random variables that the stochastic finite element module can handle.";
      throw FlxException("flxSFEM::assembleDOF_2", ssV.str() );
    }
    
  // assemble RV indices
    tuint ID = 0;
    for (size_t i=0;i<set_vec.size();++i) {
      RBRV_set_noise_RF* setRF = dynamic_cast<RBRV_set_noise_RF*>(set_vec[i]);
      if (setRF==NULL) throw FlxException_Crude("flxSFEM::assembleDOF_3");
      setRF->set_RndBox_index(ID);
      ID += setRF->get_NRV_only_this();
    }
    if (ID!=RndBox.get_NRV()) throw FlxException_Crude("flxSFEM::assembleDOF_4");
  
  // allocate the SSFEM stiffness matrix
    if (StiffnessMtx) {
      delete StiffnessMtx;
    }
    StiffnessMtx = new flxStiffnessMtxSFEM(RndBox.get_NRV(), p, this, set_vec);
}


