/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef sfem_kernel_H
#define sfem_kernel_H

#include "flxpeval.h"
#include "flxfem.h"
#include "flxrbrv_rvs.h"


#include <cstddef>


#if defined(_MSC_VER)
  #include "flxlibssfem_export.h"
#else
  #define FLXLIBSSFEM_EXPORT
#endif


  
/**
* @brief Class for Polynomial Chaos (Hermite Polynomials)
*/
class FLXLIBSSFEM_EXPORT PolyChaos {
  private:
    /**
    * @brief dimension of the Polynomial Chaos
    */
    int M;
    /**
    * @brief order of the Polynomial Chaos
    */
    int p;
    /**
    * @brief stores sequences  (qs[jA][M])
    */
    int **qS;
    /**
    * @brief stores number of sequences
    */
    tulong jA;
    /**
    * @brief Stores Coefficient of Hermite Polynomials
    */
    tdouble **H;
    /**
    * @brief Stores the squared expectation of the sequences
    */
    tdouble *ExpSq;
    
    /**
    * @brief free allocated memory
    */
    void delMem();
    /**
    * @brief extracts the integer sequence out of the ball sequence
    */
    int *ExtrSeq(const int n, int *balls) const;
    /**
    * @brief returns false if there is no new combination
    */
    const bool NxtSmp(const int n, const int m, int *balls) const;
    /**
    * @brief calculate expectation (square norm of the basis function Psi_j)
    */
    const tdouble calc_ExpPB2(const tulong j);
  public:
    /**
    * @brief creates the polynomial chaos
    * @param MV The dimension (number of random variables)
    * @param pV the maximum order of the polynomial base functions
    */
    PolyChaos(const int MV, const int pV);
    ~PolyChaos();
    /**
    * @brief set different values for Dimension (M) and order (p)
    */
    void GenerateSequence(const int MV, const int pV);
    /**
    * @brief return number of sequences (number of base functions)
    */
    const tulong get_SeqNumb() const {return jA;}
    /**
    * @brief calculate value of q.th Hermite Polynomial at position x
    */
    const tdouble calc_H1D(const int q, const tdouble x) const;
    /**
    * @brief return the q.th Hermite Polynomial for RV_m
    */
    FlxPeval* get_H1D(const int q, const int m) const;
    /**
    * @brief calculate value of j.th Polynomial at position x
    * @param j number of the polynomial (base function) to evaluate
    * @param x vector of random variables (size=Number of random variables = M)
    */
    const tdouble calc_Poly(const tulong j, const flxVec& x) const;
    /**
    * @brief return the j.th Polynomial
    */
    FlxPeval* get_Poly(const tulong j) const;
    /**
    * @brief calculate value of the realization Polynomial at position x
    * @param U parameter vector (coordinates of the base)
    * @param x vector of random variables (size=Number of random variables = M)
    */
    const tdouble calc_Realization(const flxVec& U, const flxVec& x) const;
    /**
    * @brief return the realization polynomial
    */
    FlxPeval* get_Realization(const flxVec& U) const;
    
    /**
    * @brief calculate coefficients c_ijk
    */
    const tdouble calc_c_ijk(const int i, tulong j, tulong k);
    /**
    * @brief calculate expectation (square norm of the basis function Psi_j)
    */
    const tdouble get_ExpPB2(const tulong j) const;
    /**
    * @brief get the entries of K_jk (similar to calc_c_ijk)
    */
    void calc_Kjk(tulong j, tulong k, int& i, tdouble& f);
    
    /**
    * @brief returns the dimension of the stochastic space
    */
    const int get_dim() const { return M;}
    /**
    * @brief returns the maximum order of the polynomial base functions
    */
    const int get_order() const { return p;}
};

class flxStiffnessMtxSFEM : public flxStiffnessMtxBase {
  protected:
    /**
    * @brief calss for generating the polynomial chaos
    */
    PolyChaos pc;
    /**
    * @brief Stiffness Matrix (DofAxDofA)[M+1]
    */
    FlxMtx_baseS **StfMtxV;
    
    virtual void allocate_SolutionVec();
    void delete_StfMtxV();
  public:
    RBRV_constructor RndBox;
    
    flxStiffnessMtxSFEM(const int dim, const int order, flxFEM* FEM, std::vector<RBRV_set_base*> set_vec);
    virtual ~flxStiffnessMtxSFEM();
    virtual void assemble_stf();
    virtual void assemble_forc(const bool errSerious=true);
    virtual const double solve(tdouble &eps, tnlong& iter, tuint pcn, bool startZero,const bool errSerious=true, const bool verboseOutput=true);
    virtual void output_solVec(std::ostream& sout, bool verbose);
    virtual const tdouble get_U(const tnlong dofindex) const;
    virtual void get_U(const tnlong dofindex, tdouble *v, const tuint N) const;
    virtual const tdouble get_F(const flxDOFn* Dof);
    virtual const bool condensation_is_allowed() const { return false; }
    
    /**
    * @brief returns the solution polynomial of a certain DOF and allocates memory for the polynomial
    */
    FlxPeval* get_Pol(const tnlong dofindex);
    const PolyChaos& get_pc() { return pc; }
};

class flxSFEM : public flxFEM {
  private:
    /**
    * @brief order of the stochastic space (just used to store the value for assemble DOF)
    */
    int p;
    flxStiffnessMtxSFEM* get_Sstfmtx();
  public:
    flxSFEM(int order, FlxConstantBox* constBox);
    virtual ~flxSFEM();
    virtual void assembleDOF();
    
    /**
    * @brief returns the solution polynomial of a certain DOF and allocates memory for the polynomial
    */
    FlxPeval* get_Pol(const tnlong dofindex);
    RBRV_constructor& get_RndBox();
    
    static RBRV_set_box* rbrv_box;
};





// -------------------------- inline Functions ----------------------------------------------------------------------

inline const tdouble PolyChaos::calc_H1D(const int q, const tdouble x) const
{
  #if FLX_DEBUG
    if ( q < 0 || q > p) {
      std::ostringstream ssV_2;
      ssV_2 << "q (q=" << q << "; q.th Hermite Polynomail) has to be in the Range [0;" << p << "].";
      throw FlxException("sfem::PolyChaos::calc_H1D_1", ssV_2.str() );
    }
  #endif
  int t = (q+2)/2;	// number of coefficients unequal to 0
  int u = (q) % 2;	// even number=0; uneven number=1
  tdouble y = 0;
  for (int i=0; i<t;i++) {
    y+=H[q][i]*pow(x,i*2+u);
  }
  return y;
}

inline const tdouble PolyChaos::calc_Poly(const tulong j, const flxVec& x) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA ) {
      std::ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Poly_1", ssV_2.str() );
    }
    if (x.get_N() != static_cast<tuint>(M) ) {
      std::ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Poly_2", ssV_2.str() );
    }
  #endif
  int *qSj = qS[j];
  tdouble f = ONE;
  for (register int i = 0; i < M; ++i) {
    if (qSj[i]!=0) {
      f *= calc_H1D(qSj[i],x[i]);
    }
  }
  return f;  
}






#endif // sfem_kernel_H
