/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef sfem_sfem_H
#define sfem_sfem_H

#include "flxobjfem.h"
#include "ssfem_kernel.h"
#include "flxobjrandom.h"



// -------------------------- FlxObjSFEM ----------------------------------------------------------------------

/**
* @brief object class: define a function
*
* sfem new
*/
class FlxObjSFEM_new : public FlxObjOutputBase {
  private:
    int p;
    void task();
  public:
    FlxObjSFEM_new(bool dolog,int p, std::string ostreamV, bool verbose ) :FlxObjOutputBase(dolog,ostreamV,verbose),p(p){};
};


/**
* @brief object read class: for FlxObjSFEM
*/
class FlxObjReadSFEM : public FlxObjReadControlFEM {
  private:
    virtual FlxObjBase* read_new(bool errSerious=false);
    virtual int get_order();
  public:
    FlxObjReadSFEM();
};


// -------------------------- FlxObjSFEMtest ----------------------------------------------------------------------

/**
* @brief object class: SFEM test class
*
* calc FlxFunction
*/
class FlxObjSFEMtest : public FlxObjOutputBase {
  private:
    int M,p;
    void task();
  public:
    FlxObjSFEMtest ( bool dolog, int MV, int pV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), M(MV), p(pV) {}
    //~FlxObjCalc() { delete fun; }
};

/**
* @brief object read class: for FlxObjSFEMtest
*/
class FlxObjReadSFEMtest : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};



// -------------------------- Functions ---------------------------------------------------------------------

class FunSFEM_moment : public FunFEM_base {
  private:
    FunBase* moment;
  public:
    FunSFEM_moment(FunBase* node,FlxDof::dof dofV,FunBase* moment):FunFEM_base(node,dofV),moment(moment){}
    ~FunSFEM_moment() { delete moment;}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) {return ( FunFEM_base::search_circref(fcr) || moment->search_circref(fcr));};
    const std::string write();
    const tuint precedence() { return 0; }
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic read class of FunSFEM_moment
*/
class FunReadSFEM_moment : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};


class FunSFEM_covariance : public FunFEM_base {
  private:
    FunBase* node2;
    FunBase* joint2;
    FlxDof::dof dof2;
  public:
    FunSFEM_covariance(FunBase* node,FlxDof::dof dofV,FunBase* node2,FlxDof::dof dof2):FunFEM_base(node,dofV),node2(node2),dof2(dof2){}
    ~FunSFEM_covariance() { delete node2; delete joint2;}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) {return ( FunFEM_base::search_circref(fcr) || node2->search_circref(fcr) || joint2->search_circref(fcr) );};
    const std::string write();
    const tuint precedence() { return 0; }
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic read class of FunReadSFEM_covariance
*/
class FunReadSFEM_covariance : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};


//======================= Random Variable Material ============================================

class flxIsoRVMat : public flxIsotropicMaterial, public FlxBoxBase2, public FlxDataBase {
  private:
    /**
    * @brief standard deviation
    */
    FlxFunction* sD;
    /**
    * @brief pointer to random variable in corresponding set
    */
    const tdouble* y_smp;
    /**
    * @brief pointer to set of random variables
    */
    RBRV_set_noise_RF* ts;
  public:
    flxIsoRVMat(FlxFunction* E, FlxFunction* sD, FlxFunction* nu, FlxFunction* density);
    flxIsoRVMat(const flxIsoRVMat& FlxM, const tuint mid);
    virtual ~flxIsoRVMat() { delete sD; };
    virtual flxIsoRVMat* copy(const tuint mid);
    virtual void log_info(std::ostream& lout);
    virtual const tdouble get_E() const;
    virtual const tdouble get_G() const;
    virtual const tdouble get_lambda() const;
};

/**
* @brief object read class: for FlxObjIsoMaterial (random material - variable)
*/
class FlxObjReadIsoRVmat : public FlxObjReadIsoMaterial {
  private:
    FlxFunction* read_matprop(char &c,bool errSerious=true);
  protected:
    void get_matprop(FlxFunctionPtr& E, FlxFunctionPtr& sD, FlxFunctionPtr& nu, bool errSerious=true);
  public:
    virtual ~FlxObjReadIsoRVmat() {};
    FlxObjBase* read ();
};


//======================= Random Field Material ============================================
  
class flxIsoRFMat : public flxIsotropicMaterial, public FlxBoxBase2 {
  private:
    FlxRandomField_base* RFptr;
  public:
    flxIsoRFMat(FlxRandomField_base* RFptr, FlxFunction* nu, FlxFunction* density, const tuint deg_E);
    flxIsoRFMat(const flxIsoRFMat& FlxM, const tuint mid);
    virtual ~flxIsoRFMat() {};
    virtual flxIsoRFMat* copy(const tuint mid);
    virtual void log_info(std::ostream& lout);
    const bool depend_E_on_Const(const tdouble* const thenumber);
    virtual const tdouble get_E() const;
    virtual const tdouble get_G() const;
    virtual const tdouble get_lambda() const;
};

/**
* @brief object read class: for FlxObjIsoMaterial (random material - field)
*/
class FlxObjReadIsoRFmat : public FlxObjReadIsoRVmat {
  public:
    FlxObjBase* read ();
};




#endif // sfem_sfem_H
